#define TINY_GSM_MODEM_SIM7600
#include <HardwareSerial.h>
HardwareSerial SerialAT(1);
HardwareSerial SerialSensor(2);

#include <MAGELLAN_MQTT.h>
#include <TinyGsmClient.h>
TinyGsm modem(SerialAT);
TinyGsmClient gsmclient(modem);
MAGELLAN_MQTT magelclient(gsmclient);

// Dependencies

#include "config.hpp"
#include "lcd.hpp"
#include "eeprom.hpp"
#include "mcp23017.hpp"
#include "ntptime.hpp"
#include "rtctime.hpp"
#include "sensor.hpp"

// Prototype gsm
void toggleLED(void *parameter);

void debug_GSM();
int8_t sendATcommand(char *ATcommand, char *expected_answer,
                     unsigned int timeout);
void initModem();
void modemPowerOn();
// Prototype magel
void initMagel();
void handleControl(JsonObject docObject);
void magel_loop();

/*
 * Setup function
 */
void setup()
{
  Serial.begin(115200);
  SerialAT.begin(115200, SERIAL_8N1, RXD1, TXD1);
  SerialSensor.begin(9600, SERIAL_8N1, RXD2, TXD2);

  delay(6000);
  pinMode(GSM_485_SELECTOR, OUTPUT);
  digitalWrite(GSM_485_SELECTOR, LOW);

  // i2c
  // scani2c();
  autoconfiglcd();  // 1
  rtc_setup(false); // 2
  rtc_showtime();   // 3

  initLcd();
  setupLcdIcon();

  initGPIO();   // 4
  mcp0_begin(); // 5

  lcd_showtime(); // 6
  // testSensor();   // 7

  // Task
  xTaskCreate(toggleLED, "toggle LED", 1000, NULL, 3, NULL);
  xTaskCreate(heartbeat, "Heart Beat", 2048, NULL, 2, NULL);

  // Turn on Modem
  oneBeeb();
  connectSim();
  EnableModem(); // 8
  // wait uart stable
  Serial.println("Wait 15 Sec ... ");
  delay(15000);
  Serial.print("MODEM_STATUS : ");
  Serial.println(digitalRead(MODEM_STATUS));
  // Initial gsm
  initModem(); // 9
  twoBeep();
  initMagel(); // 10

} // end of setup

/*
 * Setup function
 */

void loop()
{
  if (!modem.isGprsConnected())
  {
    Serial.println("Reconnect Modem");
    modem.gprsConnect(apn, gprsUser, gprsPass); // TODO process return val
  }

  unsigned long currentMillis = millis();
  // if magellen lost connection
  if ((magelclient.isConnected() == false) &&
      (currentMillis - magel_previousMillis >= magel_interval))
  {
    Serial.println("[Debug] Loop  try to reconnect to Magellance");
    // magelclient.begin(setting);
    // delay(2000);
    magelclient.reconnect();
    delay(1000);
    // reset
    magel_previousMillis = currentMillis;
  }
  else
  {
    magel_loop();
  }
  if (awakenByInterrupt)
  {
    handleInterrupt();
  }
  // Display & Report Data
  if (currentMillis - data_previousMillis >= data_interval)
  {
    // getData
    clearLCDLine(1);
    clearLCDLine(2);
    // get Light
    Serial.println("## Real Soil Sensor");
    node3.init();
    // node3.getString();
    SoilHumidData = node3.getSoilHumid();
    SoilTempData = node3.getSoilTemp();

    magelclient.control.request();

    Serial.println("## Read Light ");
    myLux.begin();
    Light = (int)myLux.lightStrengthLux();
    Serial.print("Light : ");
    Serial.println(Light);
    if (Light <= 0)
    {
      Light = 0;
    }
    Serial.print("Light : ");
    Serial.println(Light);
    magelclient.control.request();
    initSHT();
    Serial.println("## Read SHT ");
    float t = sht31.readTemperature();
    float h = sht31.readHumidity();

    if (!isnan(t))
    { // check if 'is not a number'
      Serial.print("Temp *C = ");
      Serial.print(t);
      Serial.print("\t\t");
    }
    else
    {
      Serial.println("Failed to read temperature");
    }

    if (!isnan(h))
    { // check if 'is not a number'
      Serial.print("Hum. % = ");
      Serial.println(h);
    }
    else
    {
      Serial.println("Failed to read humidity");
    }
    Serial.println("## Show LCD Data ");
    iconTemp();
    iconHumid();
    iconAir();
    iconLight();
    lcd.setCursor(0, 0);
    lcd.print("Read KMM Sensor 485");
    showLedData(3, 1, SoilTempData);
    showLedData(13, 1, SoilHumidData);

    showLedData(3, 2, t);
    showLedData(13, 2, h);

    showLedData(3, 3, Light);
    magelclient.control.request();
    delay(1000);
    
    magelclient.sensor.add("Soil", SoilHumidData);
    magelclient.sensor.add("SoilTemp", SoilTempData);
    magelclient.sensor.add("Light", Light);
    magelclient.sensor.add("AirTemp", t);
    magelclient.sensor.add("AirHumid", h);
    magelclient.sensor.report();
    data_previousMillis = currentMillis;
  }

  lcd_showtime();

} // end of loop

/*
 * Toggle LED
 */
void toggleLED(void *parameter)
{
  for (;;)
  { // infinite loop

    // Turn the LED on
    digitalWrite(LED_BUILTIN, HIGH);

    // Pause the task for 500ms
    vTaskDelay(500 / portTICK_PERIOD_MS);

    // Turn the LED off
    digitalWrite(LED_BUILTIN, LOW);

    // Pause the task again for 500ms
    vTaskDelay(500 / portTICK_PERIOD_MS);
  }
}

/*
 * HeartBeat
 */
void heartbeat(void *parameter)
{
  for (;;)
  { // infinite loop
    Serial.println("Start Heart Beat");
    // Turn the hearbeat
    iconheartbeat();
    digitalWrite(HEARTBEAT, LOW);
    digitalWrite(HEARTBEAT, HIGH);

    //  Hold 1 Sec
    vTaskDelay(2000 / portTICK_PERIOD_MS);
    cleariconheartbeat();
    // Turn the heartbeat
    digitalWrite(HEARTBEAT, LOW);

    // Pause the task again for 60 Sec
    vTaskDelay(10000 / portTICK_PERIOD_MS);
  }
}

void hearbeatManual()
{
  iconheartbeat();
  digitalWrite(HEARTBEAT, LOW);
  digitalWrite(HEARTBEAT, HIGH);

  //  Hold 1 Sec
  vTaskDelay(2000 / portTICK_PERIOD_MS);
  cleariconheartbeat();
  // Turn the heartbeat
  digitalWrite(HEARTBEAT, LOW);
}

void getDataTask(void *parameter)
{
  for (;;)
  { // infinite loop
    getData();
    // Pause the task again for 500ms
    vTaskDelay(180000 / portTICK_PERIOD_MS);
  }
}

void testSensor()
{
  // getData
  clearLCDLine(1);
  clearLCDLine(2);
  // get Light
  Serial.println("## 7 testSensor ");
  node3.init();
  // node3.getString();
  SoilHumidData = node3.getSoilHumid();
  SoilTempData = node3.getSoilTemp();

  Serial.println("## Read Light ");
  myLux.begin();
  Light = (int)myLux.lightStrengthLux();
  Serial.print("Light : ");
  Serial.println(Light);
  if (Light <= 0)
  {
    Light = 0;
  }
  Serial.print("Light : ");
  Serial.println(Light);

  initSHT();
  Serial.println("## Read SHT ");
  float t = sht31.readTemperature();
  float h = sht31.readHumidity();

  if (!isnan(t))
  { // check if 'is not a number'
    Serial.print("Temp *C = ");
    Serial.print(t);
    Serial.print("\t\t");
  }
  else
  {
    Serial.println("Failed to read temperature");
  }

  if (!isnan(h))
  { // check if 'is not a number'
    Serial.print("Hum. % = ");
    Serial.println(h);
  }
  else
  {
    Serial.println("Failed to read humidity");
  }
  Serial.println("## Show LCD Data ");
  iconTemp();
  iconHumid();
  iconAir();
  iconLight();
  lcd.setCursor(0, 0);
  lcd.print("Read KMM Sensor 485");
  showLedData(3, 1, SoilTempData);
  showLedData(13, 1, SoilHumidData);

  showLedData(3, 2, t);
  showLedData(13, 2, h);

  showLedData(3, 3, Light);

  delay(5000);
}

void oneBeeb()
{
  digitalWrite(BUZZER, LOW);
  delay(1000);
  digitalWrite(BUZZER, HIGH);
  delay(1000);
}

void twoBeep()
{
  oneBeeb();
  oneBeeb();
}

void threeBeep()
{
  oneBeeb();
  oneBeeb();
  oneBeeb();
}

void initGPIO()
{
  Serial.println("4 initGPIO ");
  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(POWER_ON_SENSER, OUTPUT);
  digitalWrite(POWER_ON_SENSER, LOW); // shutdown power sensor

  // Set Interupt pin
  pinMode(ISR_PIN, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(ISR_PIN), intCallBack, FALLING);

  // Set input output
  pinMode(HEARTBEAT, OUTPUT);
  pinMode(MODEM_POWER_ON, OUTPUT);
  pinMode(MODEM_KEY_ON, OUTPUT);
  pinMode(MODEM_STATUS, INPUT);
  digitalWrite(HEARTBEAT, LOW);

  pinMode(BUZZER, OUTPUT);
  digitalWrite(BUZZER, HIGH);

  RELAY1_OFF;
  RELAY2_OFF;
  RELAY3_OFF;
  RELAY4_OFF;
}

void testGPIO()
{
  Serial.print("Test GPIO ");
  RELAY1_ON;
  delay(2000);
  RELAY1_OFF;
  delay(2000);
  RELAY2_ON;
  delay(2000);
  RELAY2_OFF;
  delay(2000);
  RELAY3_ON;
  delay(2000);
  RELAY3_OFF;
  delay(2000);
  RELAY4_ON;
  delay(2000);
  RELAY4_OFF;
  delay(2000);
  LED3_ON;
  delay(2000);
  LED3_OFF;
  delay(2000);
  LED2_ON;
  delay(2000);
  LED2_OFF;
  delay(2000);
  LED4G_ON;
  delay(2000);
  LED4G_OFF;
  delay(2000);
  LEDWIFI_ON;
  delay(2000);
  LEDWIFI_OFF;
  delay(2000);
}

void initModem()
{

  Serial.println("9 DONE Modem is ready!");
  Serial.println("Restart modem...");
  Serial.println("Initializing modem...");

  if (!modem.init())
  {
    Serial.println(
        "Failed to restart modem, attempting to continue without restarting");
  }

  Serial.print("Connecting to APN: ");
  Serial.print(apn);
  if (!modem.gprsConnect(apn, gprsUser, gprsPass))
  {
    Serial.println(" fail");
    connectFail();
  }
  else
  {
    Serial.println(" OK");
    connectSuccess();
  }

  debug_GSM();
}

void PowerOnModem()
{
  Serial.println("Power On Modem ");
  pinMode(MODEM_POWER_ON, OUTPUT);
  Serial.println("MODEM_POWER_ON ");
  digitalWrite(MODEM_POWER_ON, HIGH);
  delay(3000);
}

void EnableModem()
{
  Serial.println("8 Turn on Modem");
  PowerOnModem();
  delay(3000);
  // check press on , press off
  Serial.println("MODEM_KEY_ON ");
  Serial.print("MODEM_STATUS : ");
  Serial.println(digitalRead(MODEM_STATUS));
  if (digitalRead(MODEM_STATUS) == 1)
  {
    Serial.println("Press ...");
    digitalWrite(MODEM_KEY_ON, HIGH);
    delay(5000);
    digitalWrite(MODEM_KEY_ON, LOW);
  }
  else
  {
    Serial.println("Not Press ...");
  }
}
void modemPowerOff()
{
  Serial.println("Power Off Modem ");
  pinMode(MODEM_POWER_ON, OUTPUT);
  Serial.println("MODEM_POWER_ON ");
  digitalWrite(MODEM_POWER_ON, LOW);
  delay(3000);
}

void debug_GSM()
{

  Serial.println("debug_GSM()..");

  String imei = "";
  String imsi = "";
  int csq = 0;
  String ccid = "";
  IPAddress localip;
  String modemInfo;

  modemInfo = modem.getModemInfo();
  delay(2000);
  Serial.print("Modem Info: ");
  Serial.println(modemInfo);
  ccid = modem.getSimCCID();
  // DBG("CCID:", ccid);
  Serial.print("[CCID: ] ");
  Serial.println(ccid);
  imei = modem.getIMEI();
  // DBG("IMSI:", imsi);
  Serial.print("[IMEI: module ] ");
  Serial.println(imei);

  imsi = modem.getIMSI();
  // DBG("IMSI:", imsi);
  Serial.print("[IMSI: sim] ");
  Serial.println(imsi);

  String cop = modem.getOperator();
  // DBG("Operator:", cop);
  Serial.print("[Operator: ] ");
  Serial.println(cop);

  localip = modem.localIP();
  // DBG("Local IP:", local);
  Serial.print("[Local IP: ] ");
  Serial.println(localip);
  showip(localip.toString());
  csq = modem.getSignalQuality();
  // DBG("Signal quality:", csq);
  Serial.print("[Signal quality: ] ");
  Serial.println(csq);
  Serial.println("+----------------------------------------+");
  showcsq(csq);
}

int8_t sendATcommand(char *ATcommand, char *expected_answer,
                     unsigned int timeout)
{
  Serial.println("┌------------------------------------------------┐");
  Serial.println("│ sendATcommand                                  │");
  Serial.println("└------------------------------------------------┘");

  char response[350];
  int x = 0;
  unsigned long previous;
  delay(250); // 500

  int answer = 0;
  memset(response, '\0', 350); // Initialize the string
  delay(100);
  while (SerialAT.available() > 0)
  {
    SerialAT.read(); // Clean the input buffer
  }
  Serial.print("[+]  Sending... command :  ");
  Serial.println(ATcommand);
  SerialAT.println(ATcommand); // Send the AT command
  delay(100);

  x = 0;
  previous = millis();
  // this loop waits for the answer
  do
  {
    if (SerialAT.available() != 0)
    {
      // if there are data in the UART input buffer, reads it and checks for
      // the asnwer
      response[x] = SerialAT.read();

      Serial.print(response[x]);

      x++;
      // check if the desired answer  is in the response of the module
      // Success modem  OK, ERROR
      if (strstr(response, expected_answer) != NULL)
      {
        answer = 1; // ok
        break;
      }
      else if (strstr(response, "ERROR") != NULL)
      {
        Serial.println("[-] debug AT Error");
        answer = 0; // error
        break;
      }
    }
  } while ((answer == 0) && ((millis() - previous) <
                             timeout)); // Waits for the asnwer with time out

  Serial.println("\n++ Response: ");
  Serial.println(response);
  Serial.println("\n++ Answer: ");
  Serial.println(answer);
  return answer;
}

// magellan
void initMagel()
{
  Serial.println("10 initMagel Connect magellan");
  // add config
  setting.ThingIdentifier = thingIdentifier;
  setting.ThingSecret = thingSecret;
  setting.endpoint = endpoint;
  setting.clientBufferSize = defaultOTABuffer;
  // debug
  Serial.print("ThingIdentifier : ");
  Serial.println(magelclient.info.getThingIdentifier());
  Serial.print("getThingSecret : ");
  Serial.println(magelclient.info.getThingSecret());
  Serial.print("getThingToken : ");
  Serial.println(magelclient.info.getThingToken());
  magelclient.sensor.setJSONBufferSize(2048);
  // magelclient.getControlJSON(
  //     [](String jsonStr)
  //     { Serial.println("JSON control Str: " + jsonStr); });
  // // Register callback
  magelclient.getControlJSON(handleControl);

  // connect
  magelclient.begin(setting);

  // get report success CODE = 20000
  magelclient.getResponse(RESP_REPORT_JSON, [](EVENTS events)
                          {
    Serial.print("[RESP REPORT] code: ");
    Serial.println(events.CODE);
    // follow status code on https://magellan.ais.co.th/api-document/3/0 {Error
    // code topic}
    Serial.print("# [RESP REPORT] response message: ");
    Serial.println(events.RESP); });
  if (magelclient.isConnected() == true)
  {
    connectmagelsuccess();
  }
  else
  {
    connectmagelfail();
  }
} // end of initmagel

// magel callback response
void handleControl(JsonObject docObject)
{
  serializeJson(docObject, Serial);
  Serial.println("\n# doc control size: " + String(docObject.size()));
  // buffer value form control key "Pump"
  String rly1 = docObject["Pump"];
  if (rly1.indexOf("null") == -1)
  {
    int rly1_int = rly1.toInt();
    magelclient.sensor.add("Pump", rly1_int);
    // control relay1
    if (rly1_int == 1)
    {
      RELAY1_ON;
      pump_status = 1;
      mem.writeInt(RELAY1_MEM, 1);
    }
    else
    {
      RELAY1_OFF;
      pump_status = 0;
      mem.writeInt(RELAY1_MEM, 0);
    }
    // ack to magellan
    magelclient.sensor.report();
    lcd_showcontrol();
  }
  // buffer value form control key "valve1"
  String rly2 = docObject["valve1"];
  if (rly2.indexOf("null") == -1)
  {
    int rly2_int = rly2.toInt();
    magelclient.sensor.add("valve1", rly2_int);
    // control relay1
    if (rly2_int == 1)
    {
      RELAY2_ON;
      zone1_status = 1;
      mem.writeInt(RELAY2_MEM, 1);
    }
    else
    {
      RELAY2_OFF;
      zone1_status = 0;
      mem.writeInt(RELAY2_MEM, 0);
    }
    // ack to magellan
    magelclient.sensor.report();
    lcd_showcontrol();
  }

  // buffer value form control key "valve2"
  String rly3 = docObject["valve2"];
  if (rly3.indexOf("null") == -1)
  {
    int rly3_int = rly3.toInt();
    magelclient.sensor.add("valve2", rly3_int);
    // control relay3
    if (rly3_int == 1)
    {
      RELAY3_ON;
      zone2_status = 1;
      mem.writeInt(RELAY3_MEM, 1);
    }
    else
    {
      RELAY3_OFF;
      zone2_status = 0;
      mem.writeInt(RELAY3_MEM, 0);
    }
    // ack to magellan
    magelclient.sensor.report();
    lcd_showcontrol();
  }

  // buffer value form control key "valve3"
  String rly4 = docObject["valve3"];
  if (rly4.indexOf("null") == -1)
  {
    int rly4_int = rly4.toInt();
    magelclient.sensor.add("valve1", rly4_int);
    // control relay4
    if (rly4_int == 1)
    {
      RELAY4_ON;
      zone3_status = 1;
      mem.writeInt(RELAY4_MEM, 1);
    }
    else
    {
      RELAY4_OFF;
      zone3_status = 0;
      mem.writeInt(RELAY4_MEM, 0);
    }
    // ack to magellan
    magelclient.sensor.report();
    lcd_showcontrol();
  }
} // end of handleControl()

void magel_loop()
{
  // loop
  magelclient.loop();
  // subscribe server
  magelclient.subscribes([]()
                         {
    // control content type JSON
    Serial.println("### Subscripe to magellan");
    magelclient.subscribe.control();
    magelclient.sensor.add(RLY1_name, 0);
    magelclient.sensor.add(RLY2_name, 0);
    magelclient.sensor.add(RLY3_name, 0);
    magelclient.sensor.add(RLY4_name, 0);
    magelclient.sensor.report(); });
  // time interval function inside every 10000 millis
  magelclient.interval(10, []()
                       {
    Serial.println("## Magel Control Request ");
    magelclient.control.request();
    Serial.println("## Report Status Memory ");
    unsigned int r1, r2, r3, r4;

    r1 = mem.readInt(RELAY1_MEM);
    r2 = mem.readInt(RELAY2_MEM);
    r3 = mem.readInt(RELAY3_MEM);
    r4 = mem.readInt(RELAY4_MEM);

    Serial.print("RELAY1 :");
    Serial.println(r1);
    Serial.print("RELAY2 :");
    Serial.println(r2);
    Serial.print("RELAY3 :");
    Serial.println(r3);
    Serial.print("RELAY4 :");
    Serial.println(r4);

    // Read bit
    // PUMPSTATE = getBit(data, 8);
    // ZONE1STATE = getBit(data, 9);
    // ZONE2STATE = getBit(data, 10);
    // ZONE3STATE = getBit(data, 11);
    Serial.println("## Report Status MCP");
    Serial.print("PUMPSTATE  :");
    Serial.println(PUMPSTATE);
    Serial.print("ZONE1STATE  :");
    Serial.println(ZONE1STATE);
    Serial.print("ZONE2STATE  :");
    Serial.println(ZONE2STATE);
    Serial.print("ZONE3STATE  :");
    Serial.println(ZONE3STATE); });

} // end of magel_loop()
