#ifndef MyClass_h
#define MyClass_h

float SoilHumidData = 0;
float SoilTempData = 0;

//---------- Switch Object ------------------
class switchCtl {
    bool status = 1;
    int pin = 0;

  public:
    switchCtl(int i ) {
      pin = i;
      pinMode(pin, INPUT);
    }

    int getStatus() {
      this->status = digitalRead(pin);
      return this->status;
    }
};

//----------- Soil and Temperature Sensor Object -------------
class SoilTemp
{
private:
    uint8_t _id;
    String _name;
    ModbusMaster node;
    Stream *_Serial;
    uint8_t j, result;
    uint16_t rawdata[9];
    int16_t data;
    float _factor;

    int16_t soilhumid = 0; //  10
    int16_t soiltemp = 0;  //  10
    // int16_t soilec = 0;    //  1000

    float soilhumid_float = 0; //  10
    float soiltemp_float = 0;  //  10
    //float soilec_float = 0;    //  1000

public:
    // contructor
    SoilTemp(uint8_t id, String name, Stream *Serial, float factor)
    {
        _id = id;
        _name = name;
        _Serial = Serial;
        _factor = factor;
    }

    void init()
    {
        Serial.println("┌------------------------------------------------┐");
        Serial.println("│ ** init Soil Temp                                  │");
        Serial.println("└------------------------------------------------┘");
        node.begin(_id, *_Serial);
        int n = 0;
        int num_try = 10;
        do
        {
            delay(500);
            result = node.readHoldingRegisters(0x0000, 3); // 1 st 16 bit
            delay(500);
            Serial.print(" result: ");
            Serial.println(result);
            if (result == node.ku8MBSuccess)
            {
                Serial.println("[+] Success to read");
                // get data
                for (j = 0; j < 1; j++)
                {
                    rawdata[j] = node.getResponseBuffer(j);
                    Serial.println(node.getResponseBuffer(j));
                }
                data = node.getResponseBuffer(0x00);
                soilhumid = data;
                soilhumid_float = soilhumid;

                data = node.getResponseBuffer(0x01);
                soiltemp = data;
                soiltemp_float = soiltemp;

                // data = node.getResponseBuffer(0x02);
                // soilec = data;
                // soilec_float = soilec;
                break;
            }
            else
            {
                Serial.println("[+] Fail to read Soil Temp");
                soiltemp = 0;
            }
            n++;
        } while (n < num_try);

        char buffer[100];
        if (n > 1)
        {
            sprintf(buffer, "Try  Soil Temp %d time", n);
            Serial.println(buffer);
        }
    }

    void getString()
    {
        Serial.println("┌------------------------------------------------┐");
        Serial.println("│ >> Get string                                  │");
        Serial.println("└------------------------------------------------┘");
        Serial.print("[+] Id: ");
        Serial.println(_id);
        Serial.print("[+] _name: ");
        Serial.println(_name);
        Serial.println("");
        Serial.print("soilhumid ");
        Serial.println(soilhumid);
        Serial.println(soilhumid_float / 10);

        Serial.print("soiltemp ");
        Serial.println(soiltemp);
        Serial.println(soiltemp_float / 10);

        // Serial.print("soilec");
        // Serial.println(soilec);
        // Serial.println(_factor * soilec_float / 1000);
    }

    float getSoilTemp()
    {
        return float(soiltemp_float / 10);
    }

    float getSoilHumid()
    {
        return float(soilhumid_float / 10);
    }

    // String getSoilEc()
    // {
    //     return String(_factor * soilec_float / 1000);
    // }
};

SoilTemp node3(3, "node3", 9600, 1.0f);

void getData()
{
    node3.init();
    // node3.getString();
    SoilHumidData = node3.getSoilHumid();
    SoilTempData = node3.getSoilTemp();
}


//----------- Humidity and Temperature Sensor Object -------------
class HumiTemp {
    int _id;
    String Temperature = "00.0";
    String Humidity = "00.0";
    uint8_t result;
    uint16_t data;

  public:
    HumiTemp(int i) {
      _id = i;
      getData();
    };

    String getId() {
      return String(this->_id) ;
    }

    String getTemperature() {
      getData();
      return this->Temperature ;
    }

    String getHumidity() {
      getData();
      return this->Humidity;
    }

    String getHumiTemp() {
      getData();
      return this->Humidity + "," + this->Temperature;
    }


    void getData() {
      node.begin(_id, Serial);
      // Reading sensor modbus device , RS1-0102
      result = node.readHoldingRegisters(0x0000, 2);
      for(int i=0;i<50;i++) delay(1);
//      delay(50);
      if (result == node.ku8MBSuccess) {
        // Temperature value Device ID#1
        data  = node.getResponseBuffer(0x00);
        if(data <= 1000) {
          pTemp = data;
        } else {
          data = pTemp;
        }
        this->Temperature = String(data / 10) + "." + String(data % 10);
        data  = node.getResponseBuffer(0x01);
        if(data <= 1000) {
          pHumi = data;
        } else {
          data = pHumi;
        }
        this->Humidity = String(data / 10) + "." + String(data % 10);
      } else {
        Serial.println("Temp/Humi Sensor not found... ");
        Temperature = "00.0";
        Humidity = "00.0";
      }
    }
};

//---------- Ambient Light Sensor Object ----------------
class AmbientLight {
  private:
    int _id;
    String lux = "0";
    unsigned long Ambient_Light;
    uint8_t result;
    uint16_t data;
  public:
    AmbientLight(int i) {
      this->_id = i;
      getData();
    }

    void getData() {

      node.begin(_id, Serial);
      // Reading sensor modbus device , RS1-0401
      result = node.readHoldingRegisters(0x0000, 2);
      for(int i=0;i<150;i++) delay(1);
//      delay(50);
      if (result == 0) {
        // Ambient light sensor is long data type. Double word size from register word 0 and 1
        this->Ambient_Light = ((unsigned long)(node.getResponseBuffer(0x00)) << 16) | node.getResponseBuffer(0x01);
        lux = String(Ambient_Light);
      } else {
        Serial.println("Light Sensor not found... ");
        lux = "0";
      }
    }

    String getLight() {
      getData();
      return this->lux;
    }
};

//------------- Relay Object --------------
class Relay {
  private:
    int Rstatus = 0;
    int pin;
  public:
    Relay(int i = 0) {
      this->pin = i;
      pinMode(this->pin, OUTPUT);
      digitalWrite(this->pin, HIGH);
    }

    int getStatus() {
      return this->Rstatus ;
    }

    int getPin() {
      return this->pin;
    }

    void setPin(int p) {
      pin = p;
    }

    void RelayOn() {
      digitalWrite(this->pin, LOW);
      this->Rstatus = 1;
    }

    void RelayOff() {
      digitalWrite(this->pin, HIGH);
      this->Rstatus = 0;
    }
};

#endif
